/*
 * Modification History
 *
 * 2003-July-20    Jason Rohrer
 * Created.
 *
 * 2003-July-27    Jason Rohrer
 * Added socket destruction.
 * Added host list parsing.
 * Added support for exchanging port numbers with remote hosts.
 *
 * 2003-August-7   Jason Rohrer
 * Improved handling of remote address.
 * Fixed a bug in using string after deletion.
 * Fixed a memory leak.
 * Added a duplicate message detector.
 *
 * 2003-August-11   Jason Rohrer
 * Added use of connection maintainer, generic type used to avoid include loop.
 * Added log message when tag search fails for incoming message.
 * Changed to break connection if host list receipt fails.
 * Fixed bugs in bad received message logging.
 *
 * 2003-August-12   Jason Rohrer
 * Added connection status (accept/reject) after host list.
 * Moved readStreamUpToTag into common file.
 * Moved remote port reading code into ConnectionMaintainer.
 *
 * 2003-August-14   Jason Rohrer
 * Added function for registering backrouting information in
 * OutboundChannelManager.
 *
 * 2003-September-9   Jason Rohrer
 * Removed print statement.
 *
 * 2003-October-9   Jason Rohrer
 * Added support for message limiters.
 *
 * 2003-October-12   Jason Rohrer
 * Added use of message ID tracker.
 *
 * 2003-October-17   Jason Rohrer
 * Fixed to support broadcasts.
 * Fixed to include channel argument when routing.
 *
 * 2003-November-20   Jason Rohrer
 * Added a runtime-toggled connection log.
 *
 * 2003-November-24   Jason Rohrer
 * Added support for flags.
 *
 * 2003-December-5   Jason Rohrer
 * Added support for message utility.
 *
 * 2003-December-25   Jason Rohrer
 * Improved position of message payload logging.
 *
 * 2003-December-26   Jason Rohrer
 * Fixed utility counter default.  Added better logging about utility.
 *
 * 2004-January-11   Jason Rohrer
 * Made include paths explicit to help certain compilers.
 */



#include "MUTE/layers/messageRouting/ChannelReceivingThread.h"
// safe to include in the .cpp file (to avoid an include file loop)
#include "MUTE/layers/messageRouting/ConnectionMaintainer.h"

#include "minorGems/util/log/AppLog.h"
#include "minorGems/util/stringUtils.h"
#include "minorGems/util/SettingsManager.h"

#include <stdio.h>

#include "minorGems/network/p2pParts/protocolUtils.h"



ChannelReceivingThread::ChannelReceivingThread(
    InputStream *inInputStream,
    OutboundChannel *inOutboundChannel,
    Socket *inSocket,
    char *inRemoteAddress,
    int inRemotePort,
    LocalAddressReceiver *inMessageReceiver,
    OutboundChannelManager *inOutboundChannelManager,
    void *inConnectionMaintainer,
    HostCatcher *inHostCatcher,
    MessageIDTracker *inMessageIDTracker,
    MessagePerSecondLimiter *inLimiter )
    : mLock( new MutexLock() ),
      mInputStream( inInputStream ),
      mOutboundChannel( inOutboundChannel ),
      mSocket( inSocket ),
      mRemoteAddress( stringDuplicate( inRemoteAddress ) ),
      mRemotePort( inRemotePort ),
      mReceiver( inMessageReceiver ),
      mOutboundChannelManager( inOutboundChannelManager ),
      mConnectionMaintainer( inConnectionMaintainer ),
      mHostCatcher( inHostCatcher ),
      mMessageIDTracker( inMessageIDTracker ),
      mLimiter( inLimiter ),
      mStopSignal( false ),
      mFinished( false ),
      mLoggerName( NULL ),
      mMaxUtilityCounter( 0 ) {


    mLoggerName = autoSprintf( "ChannelReceivingThread %s:%d\n",
                               mRemoteAddress, mRemotePort );

    char found;
    int maxUtilityCounterSetting =
        SettingsManager::getIntSetting( "maxMessageUtility", &found );  

    if( found ) {
        mMaxUtilityCounter = maxUtilityCounterSetting;
        }
    else {
        // default to 100
        mMaxUtilityCounter = 100;
        }
    
    this->start();
    }



ChannelReceivingThread::~ChannelReceivingThread() {
    mLock->lock();
    mStopSignal = true;
    mLock->unlock();
    
    this->join();

    delete mInputStream;
    delete mLock;
    delete [] mRemoteAddress;
    delete [] mLoggerName;
    }



char ChannelReceivingThread::isFinished() {
    mLock->lock();

    char returnValue = mFinished;

    mLock->unlock();

    return returnValue;
    }



void ChannelReceivingThread::run() {
    AppLog::info( mLoggerName, "Starting up." );

    char correctMessageReceived = true;
    
    // first, receive the incoming host list
    // read up to the first occurrence of "EndHostList", allowing 10,000
    // characters
    char *readListBuffer = readStreamUpToTag( mInputStream,
                                              "EndHostList", 10000 );
    if( readListBuffer != NULL ) {

        AppLog::trace( mLoggerName, readListBuffer );
        
        // tokenize it
        SimpleVector<char *> *tokenVector =
            tokenizeString( readListBuffer ); 

        delete [] readListBuffer;
        
        
        int numTokens = tokenVector->size();
        
        // count should be second token, so skip first two

        int i;
        for( i=2; i<numTokens-1; i+=2 ) {

            char *addressString = *( tokenVector->getElement( i ) );
            char *portString = *( tokenVector->getElement( i + 1 ) );

            int port;

            int numRead = sscanf( portString, "%d", &port );

            if( numRead == 1 ) {
                HostAddress *address = new HostAddress(
                    stringDuplicate( addressString ), port );
                mHostCatcher->addHost( address );
                delete address;
                }
            }

        for( i=0; i<numTokens; i++ ) {
            delete [] *( tokenVector->getElement(i) );
            }

        delete tokenVector;
        }
    else {
        AppLog::error( mLoggerName, "Failed to receive host list." );
        correctMessageReceived = false;
        }


    // check for an accpeted connection status
    
    char connectionAccpeted = false;
    
    if( correctMessageReceived ) {
        correctMessageReceived = false;
        
        // next, receive the connection status
        char *readStatusBuffer =
            readStreamUpToTag( mInputStream, "EndConnectionStatus", 1000 );

        if( readStatusBuffer != NULL ) {
            AppLog::trace( mLoggerName, readStatusBuffer );
            
            // tokenize it
            SimpleVector<char *> *tokenVector =
                tokenizeString( readStatusBuffer ); 

            delete [] readStatusBuffer;
        
        
            int numTokens = tokenVector->size();

            if( numTokens > 2 ) {
                // second token is status
                char *status = *( tokenVector->getElement( 1 ) );

                if( strcmp( status, "Accepted" ) == 0 ) {
                    connectionAccpeted = true;
                    correctMessageReceived = true;
                    // we are actually connected now

                    // log it, if needed
                    char found;
                    int logConnectionsFlag =
                        SettingsManager::getIntSetting( "logConnections",
                                                        &found );

                    if( found && logConnectionsFlag == 1 ) {
                        FILE *connectionLogFILE =
                            fopen( "connectionHistory.log", "a" );

                        if( connectionLogFILE != NULL ) {

                            fprintf( connectionLogFILE, "%s : %d\n",
                                     mRemoteAddress,
                                     mRemotePort );

                            fclose( connectionLogFILE );
                            }
                        }
                    
                    }
                else if( strcmp( status, "Rejected" ) == 0 ) {
                    connectionAccpeted = false;
                    correctMessageReceived = true;

                    AppLog::info( mLoggerName,
                                  "Connection rejected by remote host." );
                    }                    
                }

            for( int i=0; i<numTokens; i++ ) {
                delete [] *( tokenVector->getElement( i ) );
                }
            delete tokenVector;            
            }
        
        if( !correctMessageReceived ) {
            AppLog::error( mLoggerName,
                           "Failed to receive connection status." );
            }
        }


    
    
    // now read a series of MUTE messages 
    
    mLock->lock();
    char stopped = mStopSignal;
    mLock->unlock();

    while( !stopped && correctMessageReceived  && connectionAccpeted ) {

        // read a message
        correctMessageReceived = false;
        


        // read up to the first occurrence of "Body:"
        char *readCharBuffer = readStreamUpToTag( mInputStream,
                                                  "Body:", 5000 );
                
        if( readCharBuffer != NULL ) {
            // buffer ends with "Body:" plus terminating character

            AppLog::detail( mLoggerName, "Got header" );
            AppLog::trace( mLoggerName, readCharBuffer );

            // tokenize it
            SimpleVector<char *> *tokenVector =
                tokenizeString( readCharBuffer ); 

            char *messageID = NULL;
            char *fromAddress = NULL;
            char *toAddress = NULL;
            char *flags = NULL;
            int utilityCounter = 0;
            int payloadLength = -1;
            char *dataPayload = NULL;

            int numTokens = tokenVector->size();

            int i;
            for( i=0; i<numTokens-1; i++ ) {
                char *currentToken = *( tokenVector->getElement( i ) );
                char *nextToken = *( tokenVector->getElement( i + 1 ) );

                if( strcmp( currentToken, "UniqueID:" ) == 0 ) {
                    messageID = stringDuplicate( nextToken );
                    }
                else if( strcmp( currentToken, "From:" ) == 0 ) {
                    fromAddress = stringDuplicate( nextToken );
                    }
                else if ( strcmp( currentToken, "To:" ) == 0 ) {
                    toAddress = stringDuplicate( nextToken );
                    }
                else if ( strcmp( currentToken, "Flags:" ) == 0 ) {
                    flags = stringDuplicate( nextToken );
                    }
                else if ( strcmp( currentToken, "UtilityCounter:" ) == 0 ) {
                    sscanf( nextToken, "%d", &utilityCounter );
                    }
                else if ( strcmp( currentToken, "Length:" ) == 0 ) {
                    sscanf( nextToken, "%d", &payloadLength );
                    }
                }

            // destroy vector
            for( i=0; i<numTokens; i++ ) {
                delete [] *( tokenVector->getElement( i ) );
                }
            delete tokenVector;

            
            if( messageID != NULL &&
                fromAddress != NULL &&
                toAddress != NULL &&
                flags != NULL &&
                payloadLength > 0 ) {

                // all header items read correctly

                // make sure payload not too big
                if( payloadLength <= 32768 ) {

                    dataPayload = new char[ payloadLength + 1 ];

                    int numRead =
                        mInputStream->read( (unsigned char *)dataPayload,
                                            payloadLength );

                    // terminate payload string
                    if( numRead > 0 && numRead <= payloadLength ) {
                        dataPayload[ numRead ] = '\0';
                        }
                    else {
                        dataPayload[ payloadLength ] = '\0';
                        }
                    
                    if( numRead == payloadLength ) {                        
                        // we've already traced the header
                        AppLog::trace( mLoggerName, dataPayload );
                        
                        correctMessageReceived = true;
                        // message formatted correctly,
                        // so process it

                        // obey the limit
                        // we will block here if message rate is too high
                        mLimiter->messageTransmitted();
                        
                        
                        char fresh =
                            mMessageIDTracker->checkIfIDFresh(
                                messageID );

                        if( fresh ) {
                            AppLog::detail( mLoggerName, "Message fresh." );

                            if( strstr( flags, "FRESH_ROUTE" ) ) {
                                // clear the routing information in both
                                // directions
                                mOutboundChannelManager->
                                    clearRoutingInformation( fromAddress );
                                mOutboundChannelManager->
                                    clearRoutingInformation( toAddress );
                                }
                            
                            // add new back-routing information 
                            mOutboundChannelManager->addRoutingInformation(
                                fromAddress,
                                mOutboundChannel );

                            int generatedUtility = 0;
                            char receivedLocally =
                                mReceiver->messageReceived(
                                    fromAddress,
                                    toAddress,
                                    dataPayload,
                                    &generatedUtility );

                            
                            if( receivedLocally ) {
                                AppLog::detail( mLoggerName,
                                                "Message consumed locally." );
                                }

                            
                            // add in our generated utility
                            utilityCounter += generatedUtility;
                            
                            // if we didn't consume message locally
                            // or if message is a broadcast
                            if( ! receivedLocally ||
                                strcmp( toAddress, "ALL" ) == 0 ) {

                                // if utility counter not too high
                                if( utilityCounter <= mMaxUtilityCounter ) {
                                    // route it
                                    mOutboundChannelManager->routeMessage(
                                        messageID,
                                        fromAddress,
                                        toAddress,
                                        flags,
                                        utilityCounter,
                                        dataPayload,
                                        mOutboundChannel );
                                    
                                    AppLog::detail( mLoggerName,
                                                    "Message routed onward." );
                                    }
                                else {
                                    AppLog::detail(
                                        mLoggerName,
                                        "Message dropped because its"
                                        " utility is too high.");
                                    }
                                }                            
                            }
                        else {
                            // else message ID expired, so ignore it
                            AppLog::detail( mLoggerName, "Message stale." );
                            }
                        }
                    else {
                        AppLog::error( mLoggerName,
                                       "Failed to read message data payload");
                        }
                    }
                }


            // clean up, whether correct message was received or not
            if( messageID != NULL ) {
                delete [] messageID;
                }
            if( fromAddress != NULL ) {
                delete [] fromAddress;
                }
            if( toAddress != NULL ) {
                delete [] toAddress;
                }
            if( flags != NULL ) {
                delete [] flags;
                }
            if( dataPayload != NULL ) {
                delete [] dataPayload;
                }

            delete [] readCharBuffer;
            }
        

        if( correctMessageReceived ) {
            AppLog::detail( mLoggerName, "Correct message received" );
            }
        else {
            AppLog::error( mLoggerName, "Bad message received" );
            }
        
        mLock->lock();
        stopped = mStopSignal;
        mLock->unlock();
        }


    // however we got here, we should remove the host from the catcher
    // (rejected our connection, broke the connection, sent a bad message)
    HostAddress *remoteHost =
        new HostAddress( stringDuplicate( mRemoteAddress ),
                         mRemotePort );
    mHostCatcher->noteHostBad( remoteHost );
    delete remoteHost;


    // let channel manager know that channel has been broken
    mOutboundChannelManager->channelBroken( mOutboundChannel );

    // cast to true type
    ConnectionMaintainer *maintainer =
        (ConnectionMaintainer *)mConnectionMaintainer;
    maintainer->connectionBroken();
    
    // destroy the socket
    delete mSocket;
    
    mLock->lock();
    mFinished = true;
    mLock->unlock();
    }



