/*
 * Modification History
 *
 * 2000-December-13		Jason Rohrer
 * Created.
 *
 * 2001-March-4		Jason Rohrer
 * Made sleep() static so it can be called by non-Thread classes.
 *
 * 2001-May-12   Jason Rohrer
 * Added comments about joining before destroying.
 *
 * 2002-March-29    Jason Rohrer
 * Added Fortify inclusion.
 *
 * 2002-August-5    Jason Rohrer
 * Made destructor virtual.
 */

#include "minorGems/common.h"



#ifndef THREAD_CLASS_INCLUDED
#define THREAD_CLASS_INCLUDED



#ifdef FORTIFY
#include "minorGems/util/development/fortify/fortify.h"
#endif



/**
 * Base class to be subclassed by all threads.
 *
 * Note:  Implementation for the functions defined here is provided
 *   separately for each platform (in the mac/ linux/ and win32/ 
 *   subdirectories).
 *
 * @author Jason Rohrer
 */ 
class Thread {

	public:
	
		Thread();		
		virtual ~Thread();
	
		
		/**
		 * Starts this Thread.
		 *
		 * Note that after starting a thread, it _must_ be joined before
		 * being destroyed to avoid memory leaks.
		 */
		void start();
		
		
		/**
		 * To be overriden by subclasses.
		 * This method will be run by the Thread after start() has been called.
		 */
		virtual void run() = 0;
		
		
		/**
		 * Blocks until this thread finishes executing its run() method.
		 *
		 * Must be called before destroying this thread, if this thread
		 * has been started.
		 */
		void join();

		
		/**
		 * Puts the current thread to sleep for a specified amount of time.
		 *
		 * Note that given a thread instance threadA, calling threadA.sleep()
		 * will put the calling thread to sleep.
		 *
		 * @param inTimeInMilliseconds the number of milliseconds to sleep.
		 */
		static void sleep( unsigned long inTimeInMilliseconds );


	private:
		
		/**
		 * Used by platform-specific implementations.
		 */		
		void *mNativeObjectPointer;
		
	};		
	
#endif
